from .geometry import *
from .contextualizer import Contextualizer
from .logging import logger
import requests
import logging


NIFI_HEALTHCHECK_PATH = "/healthcheck"
CONTEXT_OBSERVATION_SEMANTICS = "earth:Terrestrial earth:Region"

class KlabObservationNifiRequest(BaseModel):
    '''
    The Main Observation Class in Python for creating the JSON Payload passing to 
    the Observation Relay Processor through the flowfile. use the method `to_dict()`
    method to convert the observation object to an equivalent JSON.

    If using the ListenHTTP Processor in Apache Nifi, use the :class:`Client` class, and use the 
    `submit(:class:NifiKlabObservation)` method.

    '''

    def __init__(self, 
                 observationName:str=None,
                 observationSemantics:str=None,
                 asContext:bool=False,
                 space:Space=None,
                 time:Time=None,
                 dtURL:str=None,
                 id:int=None,
                 contextualizer:Contextualizer=None,
                 loglevel:str=logging.INFO):
        
        logger.debug("KLAB Nifi Observation Initialized")
        logger.info("Building the Nifi Observation")
        logger.setLevel(loglevel)

        if not observationName and asContext:
            raise KlabNifiException("Observation Name cannot be non null for Context Observations")
        else:
            if observationName:
                logger.info("Setting Name to the Observation")
                self.name = observationName
        
        if not observationSemantics:
            raise KlabNifiException("Observation Query must be made with a Semantics")

        logger.info("Setting Name and Semantics to the Observation")

        if contextualizer :
            logger.info("Setting Contextualizer to the Observation")
            self.contextualizer = contextualizer


        ##TODO: check how can we validate the semantics here without the Python Client
        ## Keeping it as it is for now
        self.semantics = observationSemantics 
        self.asContext = asContext

        if contextualizer and asContext:
            raise KlabNifiException("Contextualizer cannot be set for Context Observations")

        if self.asContext and (not space or not time or observationSemantics != CONTEXT_OBSERVATION_SEMANTICS) :
            raise KlabNifiException("Context Observations must have both Space and Time defined, "
            "and Semantics must be set to " + CONTEXT_OBSERVATION_SEMANTICS)

        if space and time :
            logger.debug("Setting Geometry")
            self.geometry = Geometry(space, time)

        if dtURL :
            logger.debug("Setting the Digital Twin URL")
            self.digitalTwin = dtURL
        else:
            raise KlabNifiException("Digital Twin URL cannot be Null for Observation Request")

        logger.info("Initial Validations Passed, Observation Payload Created")


class KlabNifiListenHTTPClient:
    '''
    Class to submit, an Observation to the Nifi ListenHTTP Processor
    Create an Observation, with :class:`NifiKlabObservation`, and use the `submit` 
    method to submit the created observation query to the ListemHTTP Processor Endpoint
    '''

    def __init__(self,
                 host:str="http://127.0.0.1",
                 port:str="3306",
                 healthport:str=None):
        
        self.host = host
        self.port = port
        self.healthport = healthport

        if self.healthport :
            self.healthCheck()
        else:
            logger.info("Health Check Port not configured, skipping healthcheck...")


    def healthCheck(self):
        resp = requests.get(self.host + ":" + self.healthport + NIFI_HEALTHCHECK_PATH)
        if resp.status_code != 200:
            raise KlabNifiException("HealthCheck failure") 
        logger.info("HealthCheck for ListenHTTP Processor successful")


    def submitObservation(self, obs:KlabObservationNifiRequest):
        logger.debug("Making a Post Request to the Nifi Listen HTTP Endpoint")

        if not obs:
            raise KlabNifiException("Observation cannot be Null")

        try:
            resp = requests.post(
                url=self.host + ":" + self.port,
                json = obs.to_dict()
            )
            
            if resp.status_code != 200:
                raise KlabNifiException(f"Error in Submitting Request, Request failed with Status Code: {resp.status_code}")

        except Exception:
            raise KlabNifiException("Submit Request to the Nifi Endpoint Failed")
        
        logger.info("Submitted an Observation to Nifi Endpoint Successfully")



from .commons import *
from shapely import wkt
from shapely.errors import WKTReadingError
from shapely.geometry import Point, LineString, Polygon
from typing import List, Union
from .logging import logger
from datetime import datetime, timezone
from .exception import *

class Space(BaseModel):
    def __init__(self, 
                 shape:Union[List[tuple[float, float]] , str], 
                 grid:str="1.km",
                 proj:str=KLAB_GEO_PROJ):

        if not shape:
            raise KlabNifiException("Shape cannot be None")

        if isinstance(shape, str):
            try:
                geom = wkt.loads(shape)
                logger.info("WKT String Validated Successfully")

            except WKTReadingError:
                raise KlabNifiException("Invalid Geometry")
        else:
            try:
                if len(shape) == 1:
                    geom = Point(shape[0])  
                elif len(shape) == 2:
                    geom = LineString(shape)
                else:
                    geom = Polygon(shape)
                
                if not geom.is_valid:
                    raise KlabNifiException("Invalid Geometry")
                
                logger.info("Geometry Validated Successfully")
            
            except Exception:
                raise KlabNifiException("Invalid Geomtry")
            
        
        self.shape = proj + " " + geom.wkt
        self.sgrid = grid
        self.proj = proj

class Time(BaseModel):

    TIME_SCALES = ["year"] ## Add to the scales here

    def __init__(self,
                 tstart:Union[datetime, int]=None, 
                 tend: Union[datetime, int]=None,
                 tscope: int=1,
                 tunit:str="year"):
        
        if isinstance(tstart, int):
            if not self.validate(tstart):
                raise KlabNifiException("Starting Timestamp is wrong")
        elif isinstance(tstart, datetime):
            tstart = int(tstart.timestamp() * 1000)
        else:
            raise KlabNifiException("Start Timestamp should either be a int or a datettime object")
        
        if isinstance(tend, int):
            if not self.validate(tend):
                raise KlabNifiException("End Timestamp is wrong")
        elif isinstance(tend, datetime):
            tend = int(tend.timestamp() * 1000)
        else:
            raise KlabNifiException("End Timestamp should either be a int or a datettime object")
        
        if tunit.lower() not in self.TIME_SCALES:
            raise KlabNifiException("Time Unit is wrong")
        
        if tstart > tend:
            raise KlabNifiException("Start Time cannot be greater than End Time")


        self.tstart = tstart
        self.tend = tend
        self.tunit = tunit
        self.tscope = tscope

    @staticmethod
    def validate(timestamp_str:int)->bool:
        '''
        validate the int timestamp
        '''
        try:
            ts_ms = int(timestamp_str)
            ts_sec = ts_ms / 1000
            datetime.utcfromtimestamp(ts_sec)
            return True
        except (ValueError, OverflowError):
            return False


class Geometry(BaseModel):
    '''
    Creates a Geometry, with Space and Time
    '''

    def __init__(self, space:Space=None, time:Time=None):

        if not space or not time:
            raise KlabNifiException("In geometry, both Spatial and Temporal Dimensions are required")
        
        self.space = space
        self.time = time





import logging

logger = logging.getLogger("klab-nifi-py")
logger.setLevel(logging.INFO) ## Default is INFO

# Check if the logger has any handlers already set up
if not logger.handlers:
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s] %(levelname)s: %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)


from klab_nifi_py import * 


space = Space(
    shape= "POLYGON((33.796 -7.086, 35.946 -7.086, 35.946 -9.41, 33.796 -9.41, 33.796 -7.086))",
)

dt_2020 = datetime(2020, 1, 1, 0, 0, 0)
dt_2021 = datetime(2021, 12, 31, 23, 59, 59)


time = Time(
    tstart=dt_2020,
    tend = dt_2021
    )

ctx = contextualizer.WCS(
    wcsIdentifier="im-data-global-geography__elevation-global-90m",
)

stacCtx = contextualizer.STAC(
    stacCollection="https://earth-search.aws.element84.com/v1/collections/sentinel-2-pre-c1-l2a", 
    stacAsset="red"
)



klabNifiObs = KlabObservationNifiRequest(
    ##space = space, 
    ##time = time,
    observationSemantics= "geography:Aspect",
    ##observationSemantics= "earth:Terrestrial earth:Region",
    ##asContext=True,
    ##observationName="el_capital",
    dtURL="https://services.integratedmodelling.org/runtime/main/dt/ESA_INSTITUTIONAL.fctptn9set",
    contextualizer=ctx,
)

print (klabNifiObs.to_dict())

print (klabNifiObs.to_json())

nifiklabClient = KlabNifiListenHTTPClient(port="3306", healthport="3307")
nifiklabClient.submitObservation(klabNifiObs)




from dataclasses import dataclass, field
from .commons import BaseModel


class Contextualizer(BaseModel):
    pass


@dataclass
class WCS(Contextualizer):
    serviceType: str = field(default="wcs", init=False)
    wcsIdentifier: str
    band: int = 0 ## Handling Single Band by Default
    wcsVersion: str = "2.0.1" ## Default WCS Version
    serviceUrl: str = "https://integratedmodelling.org/geoserver/ows" ## Referring to the IM GeoServer by default

    def __post_init__(self):
        self.serviceType = "wcs"
    


@dataclass
class STAC(Contextualizer):
    serviceType: str = field(default="stac", init=False)
    stacCollection: str
    stacAsset: str

    def __post_init__(self):
        self.serviceType = "stac"




from setuptools import setup, find_packages

setup(
    name="klab_nifi_py",            
    version="0.0.2",                
    author="Arnab Moitra",
    author_email="arnab.moitra@bc3research.org",
    maintainer="Artificial Intelligence For Environment and Sustainability (ARIES) Team, Basque Centre for Climate Change (BC3)",
    maintainer_email="support@integratedmodelling.org",
    description="Python Based Workflow to Interact with K.LAB Semantic Web, using Apache Nifi",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/integratedmodelling/klab-nifi",  
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3", 
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=[
        "shapely>=2.0.7"
    ],
)




import json

KLAB_GEO_PROJ = "EPSG:4326"
KLAB_UNRESOLVED_OBS_ID = -1


class BaseModel:
    def to_dict(self):
        result = {}
        for key, value in self.__dict__.items():
            if hasattr(value, "to_dict"):
                result[key] = value.to_dict()
            elif isinstance(value, list):
                result[key] = [
                    item.to_dict() if hasattr(item, "to_dict") else item
                    for item in value
                ]
            else:
                result[key] = value
        return result
    
    def to_json(self):
        return json.dumps(self.to_dict())
